//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Parameters for DDR3 controller
//
// States of DDR3 main FSM
`define DDR_IDLE        4'b0000             // Idle state
`define DDR_ACT         4'b0001             // Bank active
`define DDR_WRAP        4'b0010             // Write data with auto precharge
`define DDR_REF         4'b0100             // Refresh
`define DDR_RDAP        4'b0101             // Read data with auto precharge
`define DDR_ACT_P       4'b0110             // Bank activation is in progress
`define DDR_REF1        4'b1001             // Second stage of refresh cycle

// Commands for the main FSM
`define CMD_NOP         4'b0000             // Idle state


// States of DDR init FSM
`define DDR_I_IDLE      4'b0000             // Idle state
`define DDR_I_WAIT      4'b0010             // After reset is set to high need to wait for 500 uS more with CKE low
`define DDR_I_NOP       4'b0011             // Issue NOP command
`define DDR_I_MRS1      4'b0100             // MRS1 command
`define DDR_I_MRS2      4'b0101             // MRS2 command
`define DDR_I_MRS3      4'b0110             // MRS3 command
`define DDR_I_MRS4      4'b0111             // MRS4 command
`define DDR_I_ZQCL      4'b1000             // ZQCL command
`define DDR_I_DONE      4'b1001             // Initialization is complete  
`define DDR_I_CLB       4'b1010             // Calibration of I/O delay blocks

// States of DDR WB FSM
`define DDR_WB_IDLE     4'b0000             // Idle state
`define DDR_WB_READ     4'b0001             // Read cycle initiated by the external master
`define DDR_WB_WRITE    4'b0010             // Write cycle initiated by the external master
`define DDR_WB_READ_H   4'b0011             // Read Hit - requested data is available in the internal buffer and can be directly sent
`define DDR_WB_READ_M   4'b0100             // Read Miss - requested data is not available and needs to be read from DDR memory first
`define DDR_WB_WR_WAIT  4'b0101             // Data to be recorded but no space is available in the buffer - need to wait
`define DDR_WB_WR_DONE  4'b0110             // One packet of data has been recorded to the internal buffer 	  
`define DDR_WB_IDELAY   4'b0111             // State to change input delay if IODELAY module

// Timings for DDR init FSM - all timings in the counter clocks. At 83.347MHz one count takes 12nS. Counts accumulated
`define I_WAIT_T        16'h4200                // Wait for 200uS - 411Ah  counts (rounded to 4200h) 
`define I_NOP_T         16'hE500                // Wait for additional 500uS - A2C2h  counts (rounded to A300h)
`define I_MRS1_T        16'hE50B                //  Wait for tXPR = 120nS min (0Bh)
`define I_MRS2_T        16'hE50E                // Min tMRD = 4CK
`define I_MRS3_T        16'hE511                // Min tMRD = 4CK
`define I_MRS4_T        16'hE514                // Min tMRD = 4CK
`define I_ZQCL_T        16'hE51A                // tMOD - min greater of 12CK
`define I_DONE_T        16'hE5A0                // tZQinit - min = 512CK (128CK)

// Timings for the main FSM for refresh counter (ddr_count)
`define T_ZERO           16'h0000          // Counter zeroed
`define T_REFI           16'h0260          // tREFI = 7.8 uS with max temperature 85 degC, at 500MHz it's equal to 20'h00F3C counts - 20'h00F3C replaced by 20'h00200 for test
                                           // ddr_count works at 83MHz, at this frequency 7.8uS = 287h

// Timings for the main FSM for operations counter (op_count)
`define T_RP            8'h02               // min 13.125ns, at 100 MHz 2 cycles - 20 ns
`define T_RFC           8'h0B               // min 110ns, at 100 MHz 11 (8'hB) cycles - 110ns
`define T_WRAP          8'h05               // Min time before the next ACTIVATE command can be issued after WRAP command. T= tWL + tTW + tWR + tRP;
                                            // tWL = AL (Additive Latency) + CWL (CAS Write Latency) = 0 + 6CK
                                            // tTW = 4CK - duration of the write cycle. tTW + tWL at 100MHz will be 2 clocks
                                            // tWR(write recovery time) = min 15ns (2CK @100MHz)
                                            // tRP (precharge command period) = min 13.125ns (1CK at 100 MHz including above rounding)
                                            // So T = 2CK + 2CK + 1CK = 5CK 
`define T_TWRAP         8'h04               // Time when WRAP command is issued required to actually write one burst. T_TWRAP = tWL + tTW =  2 
`define T_RDAP          8'h0B               // Min time before the next ACTIVATE command can be issued is T=tRAS+tRP, but because tRP=tRCD and not counted
                                            // by the counter in the T_RDAP state min time will be tRAS=37.5ns (4 cycles)
                                            // Due to some time required to copy data to WB buffer this time has been extended
`define T_RCD           8'h02               // tRCDmin = 13.125nS, at 100MHz it's 16'h2

// Opcodes for signals RAS#, CAS#, WE#
`define OP_NOP          3'b111
`define OP_MRS          3'b000
`define OP_ZQCL         3'b110
`define OP_REF          3'b001
`define OP_PRE          3'b010
`define OP_ACT          3'b011
`define OP_READ         3'b101           // The same for read and read with auto precharge
`define OP_WRITE        3'b100           // The same for write and write with auto precharge

// MRS settings during initialization
`define IA_MRS0         13'b0001100110000           // a[2:0] = 3'b000; Burst length - Fixed BL8
                                                    // a[3] = 1'b0; Burst type - Sequential (nibble)
                                                    // a[6:4] = 3'b011; CAS Latency = 7 (for 187E chip)
                                                    // a[8] = 1'b1; DLL Reset - activate DLL RESET function
                                                    // a[11:9] = 3'b001; Write recovery = 5
                                                    // a[12] = 1'b0; Precharge PD (Power Down)
`define IA_MRS1         13'b0000000000010           // a[0] = 1'b0; DLL enable (normal mode)
                                                    // {a[5],a[1]} = 2'b01; ODS (Output Drive Strength), RZQ/7 (34 Ohms)
                                                    // a[4:3] = 2'b00; AL (Additive Latency) - disabled
                                                    // a[7] = 1'b0; WL (Write Levelization) = 1'b0 Disable (normal);
                                                    // {a[9],a[6],a[2]} = 3'b000; RTT_nom disabled;
                                                    // a[11] = 1'b0; TDQS (Termination Data Strobe); 0 - Disabled
                                                    // a[12] = 1'b0; Qoff - Output enable/disable (0 - enable)
`define IA_MRS2         13'b0000001001000           // a[5:3] = 3'b000; CWL (CAS Write Latency); CWL = 5CK (for clock >= 2.5ns) 
                                                    // a[6] = 1'b1; ASR (Auto Self Refresh); 1 - automatic
                                                    // a[7] = 1'b0; Self Refresh Temperature; 0 - Normal 
                                                    // a[10:9] = 2'b00; Rtt_wr; 2'b00 - Rtt_wr disabled
`define IA_MRS3         13'b0000000000000           // a[2] = 1'b0; MPR (Multi Purpose Register); 0 - Normal DRAM operations; 1 - Dataflow from MPR

`define ACT_TO          16'h100                     // Maximum idle time in the ACT state

`define RD_START       4'h8                        // Time to start data acquisition for positive cycles within read state
`define WR_START       4'h4                        // Time to start write data for negative cycles within write state