//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Top module of DDR3 SDRAM controller
//
// Controller operates at two frequencies:
// clk_div_i = 83 MHz, main frequency, used for FSM
// pll_low_clk_i = 167 MHz, frequency used to communicate with I/O blocks (ISERDES and OSERDES)
//

`include "ddr3_sdram/ddr3_parms.v"

module ddr3_top (
  clk_i, rst_i, clk_div_i, pll_low_clk_i,
  // Interface with memory chip
  dq_i, io_busy_i, 
  odt_o, reset_n_o, cke_o, adr_o, 
  ras_n_o, cas_n_o, we_n_o, ba_o, 
  dq_o, mask_o, dq_drv_o, io_clb_o, io_rst_o,                                              
  // Interface with WB
  wb_adr_i, wb_data_i, wb_we_i, wb_sel_i, wb_stb_i, wb_cyc_i,
  wb_data_o, wb_ack_o,
  // Test I/O										   
  test1_o, test2_o, test3_o
  );
  input clk_i;                  // Input clock 33 MHz
  input rst_i;                  // Input reset signal
  input clk_div_i;              // Divided DDR3 frequency
  input pll_low_clk_i;          // I/O serdes frequency divided by 4
// Interface with DDR3
  input [63:0] dq_i;             // Deserialized data input from DDR3
  input io_busy_i;               // I/O delay busy signal
  output odt_o;                  // On Die Termination
  output reset_n_o;              // Reset memory - low active
  output cke_o;                  // Clock enable
  output [12:0] adr_o;           // Address
  output ras_n_o;                // RAS strobe
  output cas_n_o;                // CAS strobe
  output we_n_o;                 // Write enable
  output [2:0] ba_o;             // Bank address
  output [63:0] dq_o;            // Data for DDR3 serializer
  output [7:0] mask_o;           // Output mask data
  output dq_drv_o;               // When set to one FSM is in write state thus DQ lines drive data outputs
  output io_clb_o;               // Calibration of I/O delays
  output io_rst_o;               // Reset of I/O delay blocks
// Wishbone interface
  input [31:0] wb_adr_i;         // Input address
  input [31:0] wb_data_i;        // Input data
  input wb_we_i;                 // Write enable signal
  input [3:0] wb_sel_i;          // Byte select signal
  input wb_stb_i;                // Input strobe
  input wb_cyc_i;                // Input cycle signal
  output [31:0] wb_data_o;       // Output data
  output wb_ack_o;               // Output acknowledge signal
// Test I/O		   
  output [3:0] test1_o;          // Test output
  output [3:0] test2_o;          // Second test output
  output [3:0] test3_o;          // Third test output
  
  reg odt_o = 1'b0;
  reg dq_drv_o;  
  
// Internal wires and registers
  reg [15:0] ddr_count;       // Counter to count time during init routine. During main ops used to count refresh time
  reg [7:0] op_count;         // Counter to count duration of commands during main operations
  reg [3:0] rd_count_pos;     // Counter used in read states - incremented on the posedge
  reg [3:0] wr_count_neg;     // Counter used in write states - incremented on the negedge
  wire [3:0] ddr_state_w;
  wire count_reset;
  
  reg ref_req;                // Refresh required
  reg wr_done_r;
  reg rd_done_r;
  
  wire rd_w;
  wire wr_w;							
  wire rd_clb_w;
  wire [9:0] c_adr_w;
  wire [12:0] r_adr_w;
  wire [2:0] ba_w;
  reg [31:0] ddr_data_r;
  
  reg [31:0] bufr [0:3];               // Buffer to store burst data acquired from DDR3 memory
  
  wire [2:0] opcodem_w;
  wire [12:0] adrm_w;
  wire [2:0] bam_w;
  wire [2:0] opcodei_w;
  wire [12:0] adri_w;
  wire [2:0] bai_w;
							
  
  wire init_done_w;							
  wire [3:0] init_state_w;
							
  wire count_resetm_w;		
  wire [3:0] wb_test_w;
  wire [31:0] adr_w;			  
  wire rdap_w;
  wire wrap_w;
  reg wdata_r;
  wire [63:0] data_port_w;
  wire [7:0] mask_w;

  assign dq_o = data_port_w;
  assign mask_o = mask_w;
  
  assign {ras_n_o, cas_n_o, we_n_o} = (init_done_w) ? opcodem_w : opcodei_w;
  assign adr_o = (init_done_w) ? adrm_w : adri_w;
  assign ba_o = (init_done_w) ? bam_w : bai_w;
  assign count_reset = (init_done_w) ? count_resetm_w : 1'b1;   // Until init routine is not done main counter is kept in reset state
  
  assign test1_o =  ddr_state_w;
  assign test2_o = wb_test_w;
  assign test3_o = init_state_w;
  
// Modules instantiation
  wb_ddr3_cntr wb_ddr3 (
  .clk_i                  (clk_i), 
  .rst_i                  (rst_i),
  .clk_div_i              (clk_div_i),
  .data_i                 (ddr_data_r),
  .wr_done_i              (wr_done_r),
  .rd_done_i              (rd_done_r),
  .wdata_i                (wdata_r),
  .adr_o                  (adr_w),
  .rd_o                   (rd_w),
  .wr_o                   (wr_w),		
  .rd_clb_o               (rd_clb_w),						   
  .data_port_o            (data_port_w),
  .mask_o                 (mask_w),
  .wb_adr_i               (wb_adr_i), 
  .wb_data_i              (wb_data_i), 
  .wb_we_i                (wb_we_i), 
  .wb_sel_i               (wb_sel_i), 
  .wb_stb_i               (wb_stb_i), 
  .wb_cyc_i               (wb_cyc_i), 
  .wb_data_o              (wb_data_o), 
  .wb_ack_o               (wb_ack_o),
  .test1_o                (wb_test_w)
  );  
  
  ddr_init_fsm ddr_init (
  .clk_i                  (pll_low_clk_i),
  .clk_div_i              (clk_div_i),
  .rst_i                  (rst_i), 						 
  .io_busy_i              (io_busy_i),
  .opcode_o               (opcodei_w),
  .adr_o                  (adri_w),                                                   
  .ba_o                   (bai_w),
  .cke_o                  (cke_o),
  .reset_n_o              (reset_n_o),
  .init_done_o            (init_done_w),
  .io_clb_o               (io_clb_o),
  .io_rst_o               (io_rst_o),
  .init_state_o           (init_state_w) 
  );
  
  ddr_fsm ddr_main_fsm (
  .clk_i                (pll_low_clk_i),       
  .clk_div_i            (clk_div_i),
  .rst_i                (rst_i),
  .ref_req_i            (ref_req),
  .rd_i                 (rd_w),
  .wr_i                 (wr_w),
  .rd_clb_i             (rd_clb_w),
  .r_adr_i              (r_adr_w),
  .c_adr_i              (c_adr_w),
  .ba_i                 (ba_w),
  .ddr_state_o          (ddr_state_w),
  .opcode_o             (opcodem_w),
  .adr_o                (adrm_w),
  .ba_o                 (bam_w),
  .count_reset_o        (count_resetm_w),
  .rdap_o               (rdap_w),
  .wrap_o               (wrap_w)
  );
  
  ddr_adrc adrc (
  .adr_i                (adr_w),
  .c_adr_o              (c_adr_w),
  .r_adr_o              (r_adr_w),
  .ba_o                 (ba_w)
  );
  										
 //------------------ChipScope logic-----------------------------
/*	
   wire [35:0] ila_cntr_w;					  
   wire [35:0] vio_cntr_w;
 					 
 	chipscope_ila_v1_05_a_0 first_ila (
      .CONTROL(ila_cntr_w), // INOUT BUS [35:0]
  	  .CLK(clk_div_i), // IN
  	  .TRIG0({3'b000,data_avail_w}), // IN BUS [3:0]
  	  .TRIG1({32'h0,data_wr_w}), // IN BUS [63:0]
  	  .TRIG2({1'b0,cnt_ddr_w})      // IN BUS [3:0]			 
 	);
  										
 	chipscope_icon_v1_06_a_0 first_icon (
      .CONTROL0(ila_cntr_w) // INOUT BUS [35:0]
  	);
*/  										
//------------End of ChipScope logic----------------------------
  										
  
  
  
  always @ (negedge clk_div_i)
    begin
      if (!rst_i)
        begin
          ddr_count <= 16'h0;	
          ref_req <= 1'b0;  
          op_count <= 8'h0;
        end
      else
        begin                   
          if (count_reset)        // Main counter logic
            ddr_count <= `T_ZERO;
          else
            ddr_count <= ddr_count + 1'b1;
            
          if ((ddr_count == `T_REFI) && init_done_w)   // Flag to define refresh requirements
            ref_req <= 1'b1;
          else if (ddr_count == `T_ZERO)
            ref_req <= 1'b0;
            
          case (ddr_state_w)                // Synthesis full case
            `DDR_IDLE, `DDR_ACT: op_count <= 8'h0;
            default: op_count <= op_count + 1'b1;
          endcase
        end
    end

  always @ (posedge pll_low_clk_i)        // posedge pll_low_clk_i
    begin	  
      if (rdap_w)     
        begin
            if (rd_count_pos != 4'hF)
              rd_count_pos <= rd_count_pos + 1'b1;
              
            if (rd_count_pos == `RD_START)
              {bufr[1],bufr[0]} <= dq_i;
            else if (rd_count_pos == (`RD_START + 1))
              begin
               {bufr[3],bufr[2]} <= dq_i;
               rd_done_r <= 1'b1;
              end
        end
      else
        begin
          rd_count_pos <= 4'h0; 
          rd_done_r <= 1'b0;
        end
    end    
    
  
  always @ (posedge clk_i)
    begin
      ddr_data_r <= bufr[wb_adr_i[3:2]];
    end
      
    
  always @ (negedge pll_low_clk_i)        // posedge pll_low_clk_i                
    begin
      if (!rst_i)
        begin
          wr_done_r <= 1'b0;							
          dq_drv_o <= 1'b1;
          wdata_r <= 1'b0;
        end
      else
        begin
          if (wrap_w)
            begin			
              dq_drv_o <= 1'b0;
              if (wr_count_neg != 4'hF)
                wr_count_neg <= wr_count_neg + 1'b1;                  
                  
              if (wr_count_neg == `WR_START)
                wdata_r <= 1'b1;
              else
                wdata_r <= 1'b0;
                  
              if (wr_count_neg == (`WR_START + 1))
                begin
                  wr_done_r <= 1'b1;
                end                
            end
          else
            begin
                wr_count_neg <= 4'h0;
                wr_done_r <= 1'b0;	
                dq_drv_o <= 1'b1;
            end
        end
    end												 

endmodule