//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Main FSM of the DDR3 SDRAM controller
//
`include "ddr3_sdram/ddr3_parms.v"

module ddr_fsm (
  clk_i, clk_div_i, rst_i, ref_req_i, rd_i, wr_i, rd_clb_i,
  r_adr_i, c_adr_i, ba_i,
  ddr_state_o, opcode_o, adr_o, ba_o, count_reset_o, rdap_o, wrap_o
  );

  input clk_i;                // Input clock
  input clk_div_i;            // Input 83MHz clock
  input rst_i;                // Input reset signal
  input ref_req_i;            // Signal that refresh is required
  input rd_i;                 // Request to read one burst of data
  input wr_i;                 // Request to write one burst of data
  input rd_clb_i;             // Run calibration routine
  input [12:0] r_adr_i;       // Input row address
  input [9:0] c_adr_i;        // Input column address
  input [2:0] ba_i;           // Input bank address
  output [3:0] ddr_state_o;   // State of the FSM
  output [2:0] opcode_o;      // Output opcode for DDR3
  output [12:0] adr_o;        // Output address for DDR3
  output [2:0] ba_o;          // Output bank address for DDR3
  output count_reset_o;       // Signal to reset main counter
  output rdap_o;              // FSM is in RDAP state
  output wrap_o;              // FSM is in WRAP state
  
  reg [2:0] opcode_o;
  reg [12:0] adr_o;
  reg [2:0] ba_o;
  reg count_reset_o;
  
  // Internal wires and registers
  reg [3:0] ddr_fsm_state;
  reg [3:0] ddr_fsm_next_state;
  reg rd_flag;                // When set read request has been received
  reg wr_flag;                // When set write request has been received
  reg [15:0] act_count;       // Counter to count idle time in the ACT state
  reg act_exp;                // Signal that maximum idle time in the ACT state has been reached
  reg [1:0] next_state;
  reg cmd_out;															   
  reg cmd_out2;
  reg [7:0] op_count;
  reg [1:0] ad_count;         // Additional counter - controls cmd_out signal
  reg [1:0] ad_count2;        // Second addtional counter -  controls cmd_out2 signal
  reg ad_count_rst;
  reg ad_count2_rst;
  reg rdap_r;
  reg wrap_r;
  
  assign rdap_o = rdap_r;
  assign wrap_o = wrap_r;
  assign ddr_state_o = ddr_fsm_state;
  
  always @ (ddr_fsm_state or ref_req_i or rd_flag or wr_flag or act_exp   // Control signal asserted on the negedge
             or next_state or cmd_out or cmd_out2 or ba_i or r_adr_i or c_adr_i)
    begin
      ddr_fsm_next_state = `DDR_IDLE;
      opcode_o = `OP_NOP;
      adr_o = 13'h0;
      ba_o = 3'b000;
      ad_count_rst = 1'b0;
      ad_count2_rst = 1'b0;
      rdap_r = 1'b0;
      wrap_r = 1'b0;
      case (ddr_fsm_state)              // Synthesis full case
        `DDR_IDLE:
          begin
            ad_count_rst = 1'b1;
            if (ref_req_i)          // Refresh is required - the highest priority
              ddr_fsm_next_state = `DDR_REF;
            else if (rd_flag || wr_flag)
              ddr_fsm_next_state = `DDR_ACT_P;
            else
              ddr_fsm_next_state = `DDR_IDLE;
          end
        `DDR_REF:             // Refresh of the memory
          begin
            ad_count2_rst = 1'b1;
            if (next_state == 2'b10)
              ddr_fsm_next_state = `DDR_REF1;
            else
              ddr_fsm_next_state = `DDR_REF;
              
            if (cmd_out)		   
              begin
                opcode_o = `OP_PRE;
                adr_o[10] = 1'b1;
              end
          end					  
        `DDR_REF1:
          begin
            if (ref_req_i)
              ddr_fsm_next_state = `DDR_REF1;
            else
              ddr_fsm_next_state = `DDR_IDLE;
              
            if (cmd_out2)
              opcode_o = `OP_REF;
          end
        `DDR_ACT_P:       // Activation is in progress
          begin
            if (next_state == 2'b01)
              ddr_fsm_next_state = `DDR_ACT;
            else
              ddr_fsm_next_state = `DDR_ACT_P;
              
            if (cmd_out)	  
              begin
                opcode_o = `OP_ACT;
                ba_o = ba_i;
                adr_o = r_adr_i;                         
              end
          end
        `DDR_ACT:   // Memory is in activated state. Need to find out if read or write request is active
          begin
            ad_count_rst = 1'b1;
            if (rd_flag)
              ddr_fsm_next_state = `DDR_RDAP;
            else if (wr_flag)
              ddr_fsm_next_state = `DDR_WRAP;
            else
              ddr_fsm_next_state = `DDR_ACT;  
          end
        `DDR_RDAP:
          begin
            rdap_r = 1'b1;
            if (next_state == 2'b10)
              ddr_fsm_next_state = `DDR_IDLE;
            else
              ddr_fsm_next_state = `DDR_RDAP;
              
            if (cmd_out)
              begin
                opcode_o = `OP_READ;
                adr_o[9:0] = c_adr_i;
                adr_o[10] = 1'b1;            // Auto precharge parameter
              end
          end  
        `DDR_WRAP:
          begin
            wrap_r = 1'b1;
            if (next_state == 2'b11)
              ddr_fsm_next_state = `DDR_IDLE;
            else
              ddr_fsm_next_state = `DDR_WRAP;					  
                  
           if (cmd_out)
             begin
               opcode_o = `OP_WRITE;
               adr_o[9:0] = c_adr_i;
               adr_o[10] = 1'b1;                // Auto precharge parameter
             end
          end
      endcase
    end
    
  always @ (posedge clk_div_i)        // Commands to the memory issued on the negedge thus states of FSM should be 
    begin                         // changing on the posedge
      if (!rst_i)
        begin
          ddr_fsm_state <= `DDR_IDLE;
        end
      else
        begin
          ddr_fsm_state <= ddr_fsm_next_state;
        end
    end
  
  always @ (negedge clk_div_i)
    begin
      if (!rst_i)
        begin
          op_count <= 8'h0;						
          next_state <= 2'b00;
          count_reset_o <= 1'b0;
        end
      else
        begin
          case (ddr_fsm_state)                // Synthesis full case
            `DDR_IDLE, `DDR_ACT: op_count <= 8'h0;
            default: op_count <= op_count + 1'b1;
          endcase								 
          
          case (ddr_fsm_state)                // Synthesis full case
            `DDR_IDLE: 
              begin
                next_state <= 2'b00;
                count_reset_o <= 1'b0;          
              end
            `DDR_REF: if (op_count == `T_RP) next_state <= 2'b10;
            `DDR_REF1: if (op_count == (`T_RP + `T_RFC)) count_reset_o <= 1'b1;
            `DDR_ACT_P: if (op_count == `T_RCD) next_state <= 2'b01;
            `DDR_RDAP: if (op_count == `T_RDAP) next_state <= 2'b10;
            `DDR_WRAP: if (op_count == `T_WRAP) next_state <= 2'b11;
          endcase
        end
    end
  
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          cmd_out <= 1'b0; 
          cmd_out2 <= 1'b0;
          ad_count2 <= 2'b00;
        end
      else
        begin
          if (ad_count_rst)
            begin
              ad_count <= 2'b00;
            end
          else if (ad_count != 2'b11)
            ad_count <= ad_count + 1'b1;
            
          if (ad_count2_rst)
            ad_count2 <= 2'b00;
          else if (ad_count2 != 2'b11)
            ad_count2 <= ad_count2 + 1'b1;
		
		      if (ad_count == 2'b01)
		        cmd_out <= 1'b1;
		      else
		        cmd_out <= 1'b0;
		        
		      if (ad_count2 == 2'b01)
		        cmd_out2 <= 1'b1;
		      else
		        cmd_out2 <= 1'b0;
        end
    end					  
    
  always @ (negedge clk_div_i)
    begin
      if (!rst_i)
        begin
          rd_flag <= 1'b0;
          wr_flag <= 1'b0;
          act_count <= 16'h0;
          act_exp <= 1'b0;
        end
      else
        begin
          if (rd_i && (ddr_fsm_state != `DDR_RDAP))     // rd_flag can be set in any state except from actual read state
            rd_flag <= 1'b1;                            // This flag is cleared in the read state
          else if (ddr_fsm_state == `DDR_RDAP)
            rd_flag <= 1'b0;
            
          if (wr_i && (ddr_fsm_state != `DDR_WRAP))   // wr_flag can be set in any state except from actual write state
            wr_flag <= 1'b1;                            // This flag is cleared in the write state
          else if (ddr_fsm_state == `DDR_WRAP)
            wr_flag <= 1'b0;
            
        end                
    end
endmodule

