//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// FSM to initialize memory
// 
// After power is stable, reset must be low for at least 200 uS to begin
// the initialization process
//
`include "ddr3_sdram/ddr3_parms.v"
//
module ddr_init_fsm (
  clk_i, clk_div_i, rst_i, io_busy_i, 
  opcode_o, adr_o, ba_o, cke_o, reset_n_o, init_done_o, io_clb_o, io_rst_o,
  init_state_o
  );
  
  input clk_i;                         // Input clock signal
  input clk_div_i;                     // Low frequency clock input
  input rst_i;                         // Input reset signal 
  input io_busy_i;                     // I/O delay block busy signal
  output [2:0] opcode_o;               // Output opcode for DDR3
  output [12:0] adr_o;                 // Output address
  output [2:0] ba_o;                   // Bank address
  output cke_o;                        // Clock enable signal
  output reset_n_o;                    // Reset for the DDR3
  output init_done_o;                  // Set to one when initialization is completed
  output io_clb_o;                     // Signal to calibrate I/O delay blocks
  output io_rst_o;                     // Signal to reset I/O delay blocks
  output [3:0] init_state_o;           // FSM state - for test purposes
  
  reg [2:0] opcode_o;
  reg [12:0] adr_o;
  reg [2:0] ba_o;											
  reg cke_o;
  reg reset_n_o;												 
  reg io_clb_o;
  reg io_rst_o;
  
  reg [15:0] init_count;
  reg [2:0] hspd_count;        // High speed counter - works at high frequency (clk_i)
  reg count_reset_o;
  
  // Internal registers and wires
  
  reg [3:0] ddr_init_state;
  reg [3:0] ddr_init_next_state;
  reg cmd_out;					
  reg init_done_r;
  
  reg  next_state; 
  reg count_stop;
  reg hspd_rst;
  
  assign init_done_o = init_done_r;									   
  assign init_state_o = ddr_init_state;
  
  
  always @ (ddr_init_state or next_state or cmd_out or io_busy_i)     // All input signal asserted on the negedge (except from ddr_init_state)
    begin
      ddr_init_next_state = `DDR_I_IDLE;
      opcode_o = `OP_NOP;
      adr_o = 13'h0;
      ba_o = 3'b000;	
      cke_o = 1'b1;
      reset_n_o = 1'b1;			
      count_stop = 1'b0;
      io_clb_o = 1'b0;
      io_rst_o = 1'b0;
      init_done_r = 1'b0;
      case (ddr_init_state)                 // Synthesis full case
        `DDR_I_IDLE:
          begin
            if (next_state)
              ddr_init_next_state = `DDR_I_WAIT;
            else
              ddr_init_next_state = `DDR_I_IDLE;              
            reset_n_o = 1'b0;
            cke_o = 1'b0;
          end
        `DDR_I_WAIT:
          begin
            if (~next_state)
              ddr_init_next_state = `DDR_I_NOP;
            else
              ddr_init_next_state = `DDR_I_WAIT;
            cke_o = 1'b0;
          end
        `DDR_I_NOP:
          begin
            if (next_state)
              ddr_init_next_state = `DDR_I_MRS1;
            else
              ddr_init_next_state = `DDR_I_NOP;
          end
        `DDR_I_MRS1:
          begin
            if (~next_state)
              ddr_init_next_state = `DDR_I_MRS2;
            else
              ddr_init_next_state = `DDR_I_MRS1;
              
            if (cmd_out)
              begin
                ba_o = 3'b010;            // Program MR2
                adr_o = `IA_MRS2;
                opcode_o = `OP_MRS;
              end
          end  
        `DDR_I_MRS2:
          begin
            if (next_state)
              ddr_init_next_state = `DDR_I_MRS3;
            else
              ddr_init_next_state = `DDR_I_MRS2;					   
          
            if (cmd_out)
              begin
                ba_o = 3'b011;            // Program MR3
                adr_o = `IA_MRS3;
                opcode_o = `OP_MRS;
              end 
          end
        `DDR_I_MRS3:
          begin
            if (~next_state)
              ddr_init_next_state = `DDR_I_MRS4;
            else
              ddr_init_next_state = `DDR_I_MRS3;
              
            if (cmd_out)
              begin
                ba_o = 3'b001;            // Program MR1
                adr_o = `IA_MRS1;
                opcode_o = `OP_MRS;
              end
          end
        `DDR_I_MRS4:
          begin
            if (next_state)		   
              begin
                ddr_init_next_state = `DDR_I_ZQCL;
                io_clb_o = 1'b1;
              end
            else
              ddr_init_next_state = `DDR_I_MRS4;
              
            if (cmd_out)
              begin
                ba_o = 3'b000;            // Program MR0
                adr_o = `IA_MRS0;
                opcode_o = `OP_MRS;
              end
          end
        `DDR_I_ZQCL:
          begin
            if (~next_state)
              begin
                ddr_init_next_state = `DDR_I_CLB;
                io_rst_o = 1'b1;
              end
            else
              ddr_init_next_state = `DDR_I_ZQCL;
              
            if (cmd_out)
              begin
                adr_o[10] = 1'b1;
                opcode_o = `OP_ZQCL;
              end		   
          end									
        `DDR_I_CLB:
          begin								   
            count_stop = 1'b1;					 
            if (io_busy_i)
              ddr_init_next_state = `DDR_I_CLB;
            else
              begin
                ddr_init_next_state = `DDR_I_DONE;
              end
          end
        `DDR_I_DONE:                              // Once init is completed FSM stays in this state
          begin
            ddr_init_next_state = `DDR_I_DONE;
            count_stop = 1'b1;	  
            init_done_r = 1'b1;
          end
      endcase  
    end
    
  always @ (posedge clk_div_i)
    begin
      if (!rst_i)
        begin
          ddr_init_state <= `DDR_I_IDLE;
        end
      else
        begin
          ddr_init_state <= ddr_init_next_state;
        end
    end
  
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          cmd_out <= 1'b0;
        end				  
      else
        begin
          if (hspd_rst)
            hspd_count <= 3'b000;
          else
            begin
              if (hspd_count != 3'b111)
                hspd_count <= hspd_count + 1'b1;
            end

          if (hspd_count == 3'b001)
            cmd_out <= 1'b1;
          else
            cmd_out <= 1'b0;
        end
    end
    
  always @ (negedge clk_div_i)
    begin
      if (!rst_i)
        begin
          init_count <= 16'h0;		 
          next_state <= 1'b0;
          hspd_rst <= 1'b1;
        end
      else
        begin	 
          if (~count_stop)
            init_count <= init_count + 1'b1;
          case (init_count)                 // Synthesis full case
            `I_WAIT_T:      next_state <= 1'b1;
            `I_NOP_T:       next_state <= 1'b0;
            `I_MRS1_T:  
              begin
                next_state <= 1'b1;
                hspd_rst <= 1'b1;
              end
            `I_MRS2_T:
              begin
                next_state <= 1'b0;
                hspd_rst <= 1'b1;
              end
            `I_MRS3_T:
              begin
                next_state <= 1'b1;
                hspd_rst <= 1'b1;
              end
            `I_MRS4_T:
              begin
                next_state <= 1'b0;
                hspd_rst <= 1'b1;
              end
            `I_ZQCL_T:
              begin
                next_state <= 1'b1;
                hspd_rst <= 1'b1;
              end
            `I_DONE_T:      next_state <= 1'b0;
            default: hspd_rst <= 1'b0;
          endcase
        end
    end
endmodule