//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Wishbone interface for DDR3 SDRAM controller
// Operates only in slave mode - processes incoming R/W
// requests but does not initiates bus cycles itself
//
// This block issues commands to read/write memory based on the 
// WB status
//
// DDR3 in this example programmed for the fixed BL8 burst length.
// 8 16bit words read at once
// When request is made to read a word WB controller neeeds to find out if 
// this word is read already and contained in the internal memory or 
// read cycle still to be completed
//
//  Data can be written to the DDR3 memory as 32-bit words (wb_sel_i == 4'b1111)
// 16-bit words (wb_sel_i == 4'b0011); and 8-bit words (wb_sel_i == 4'b0001).
// Other combinations of wb_sel_i signals are not allowed
//
`include "ddr3_sdram/ddr3_parms.v"

module wb_ddr3_cntr (
  clk_i, rst_i, clk_div_i,
  // Interface with internal devices
  data_i, wr_done_i, rd_done_i, wdata_i, 
  adr_o, rd_o, wr_o, rd_clb_o,
  data_port_o, mask_o,
  // Wishbone interface
  wb_adr_i, wb_data_i, wb_we_i, wb_sel_i, wb_stb_i,
  wb_cyc_i,
  wb_data_o, wb_ack_o,
  // Test I/O
  test1_o
  );
  
  input clk_i;                // Input clock
  input rst_i;                // Input reset signal
  input clk_div_i;            // Divided frequency of DDR3
  input [31:0] data_i;        // Input data from the memory
  input wr_done_i;            // Set to one when burst  is written
  input rd_done_i;            // Set to one when one burst read from DDR3 and needs to be transferred to the WB buffer
  input wdata_i;
  output [31:0] adr_o;        // Output address
  output rd_o;                // Request to read one burst
  output wr_o;                // Request to write one burst	  
  output rd_clb_o;            // Calibration of data reciept of DDR3
  output [63:0] data_port_o;
  output [7:0] mask_o;
  // Wishbone interface
  input [31:0] wb_adr_i;      //
  input [31:0] wb_data_i;     //
  input wb_we_i;              //
  input [3:0] wb_sel_i;       //
  input wb_stb_i;             //
  input wb_cyc_i;             // 
  output [31:0] wb_data_o;    //
  output wb_ack_o;            //
  // Test I/O
  output [3:0] test1_o;

  reg [31:0] wb_data_o;   
  reg wb_ack_o;
  reg rd_o;
  reg wr_o;															
  reg rd_clb_o;
  reg [31:0] adr_o;
  // Internal registers and wires
  reg data_avail;             // When set data is available
  reg [27:0] burst_adr;       // Address of the burst in the memory
  reg [31:0] data_bufw [0:3]; // Write buffer
  reg buf_full;               // Signal that one write buffer is full
  reg buf_rdy;                // Signal that write buffer is ready to accept data
  reg [7:0] neg_gp_cnt;       // General purpose counter incremented on the negedge of the clock
  reg [1:0] pos_gp_cnt1;      // General purpose counter
  reg data_ready;             // Set to one when one buffer is received over WB and needs to be 
                              // Tx'd to the DDR buffer
  reg next_state1;	
  reg [1:0] burst_cnt;
  reg burst_clear;
  reg [15:0] mask_reg;        // Mask registers - one bit corresponds byte
  
 
  assign data_port_o = (wdata_i) ? {data_bufw[1],data_bufw[0]} : {data_bufw[3],data_bufw[2]};
  assign mask_o = (wdata_i) ? mask_reg[7:0] : mask_reg[15:8];
  
//------------------ChipScope logic-----------------------------
/*
    wire [35:0] ila_cntr_w;					  
    wire [35:0] vio_cntr_w;
  						 
  chipscope_ila_v1_05_a_0 first_ila (
      .CONTROL(ila_cntr_w), // INOUT BUS [35:0]
      .CLK(test_clk_i), // IN
      .TRIG0({wdata_i,3'b000}), // IN BUS [3:0]
      .TRIG1({data_port_o}), // IN BUS [63:0]
      .TRIG2({mask_reg,4'h0})      // IN BUS [19:0]
  	 
  );
  
  chipscope_icon_v1_06_a_0 first_icon (
      .CONTROL0(ila_cntr_w) // INOUT BUS [35:0]
  ); 
*/  
  //------------End of ChipScope logic----------------------------
  
  
  
  //-----------------------------WB FSM----------------------------------------
  reg [3:0] ddr_wb_fsm_state;
  reg [3:0] ddr_wb_fsm_next_state;
  assign test1_o = ddr_wb_fsm_state;
  
  always @ (ddr_wb_fsm_state or wb_we_i or wb_cyc_i or wb_stb_i or buf_full or buf_rdy
              or data_avail  or next_state1)
    begin
      ddr_wb_fsm_next_state = `DDR_WB_IDLE;
      case (ddr_wb_fsm_state)               // Synthesis full case
        `DDR_WB_IDLE:
          begin
            if (~wb_we_i && wb_cyc_i && wb_stb_i)
              ddr_wb_fsm_next_state = `DDR_WB_READ;
            else if (wb_we_i && wb_cyc_i && wb_stb_i && buf_rdy)
              ddr_wb_fsm_next_state = `DDR_WB_WRITE;								 
            else
              ddr_wb_fsm_next_state = `DDR_WB_IDLE;
          end
        `DDR_WB_READ:   // Read cycle
          begin
            if (data_avail)
              ddr_wb_fsm_next_state = `DDR_WB_READ_H;
            else
              ddr_wb_fsm_next_state = `DDR_WB_READ_M;
          end
        `DDR_WB_READ_H:
          begin
            if (~data_avail && wb_cyc_i)  // If data becomes unavailable during this step it needs to be read from the DDR memory
              ddr_wb_fsm_next_state = `DDR_WB_READ_M;
            else if (~wb_cyc_i)
              ddr_wb_fsm_next_state = `DDR_WB_IDLE;
            else
              ddr_wb_fsm_next_state = `DDR_WB_READ_H;
          end
        `DDR_WB_READ_M:   // Data is not available and needs to be read from the DDR3
          begin 
            if (data_avail)     // Wait in this state until data burst is read
              ddr_wb_fsm_next_state = `DDR_WB_READ_H;
            else
              ddr_wb_fsm_next_state = `DDR_WB_READ_M;
          end
        `DDR_WB_WRITE:
          begin
            if (next_state1)
              ddr_wb_fsm_next_state = `DDR_WB_WR_DONE;
            else if (wb_cyc_i && buf_full)
              ddr_wb_fsm_next_state = `DDR_WB_WR_WAIT;
            else 
            ddr_wb_fsm_next_state = `DDR_WB_WRITE;
          end
        `DDR_WB_WR_WAIT:
          begin
            if (buf_full)
              ddr_wb_fsm_next_state = `DDR_WB_WR_WAIT;
            else
              ddr_wb_fsm_next_state = `DDR_WB_WRITE;
          end
        `DDR_WB_WR_DONE:
          begin
            ddr_wb_fsm_next_state = `DDR_WB_IDLE;
          end									   
      endcase
    end
  
  always @ (negedge clk_i)            // negedge clk_i
    begin
      if (!rst_i)
        begin
          ddr_wb_fsm_state <= `DDR_WB_IDLE;
        end
      else
        begin
          ddr_wb_fsm_state <= ddr_wb_fsm_next_state;
        end
    end
  //---------------------------End of WB FSM-----------------------------------
  
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          rd_o <= 1'b0;
          wr_o <= 1'b0;
        end
      else
        begin
          case (ddr_wb_fsm_state)           // Synthesis full case
            `DDR_WB_IDLE:
              begin
                rd_o <= 1'b0;
                wr_o <= 1'b0;
              end
            `DDR_WB_READ_H:
              begin
                rd_o <= 1'b0;
              end
            `DDR_WB_READ_M:
              begin
                rd_o <= 1'b1;
              end
            `DDR_WB_WRITE:
              begin
                wr_o <= 1'b0;
                pos_gp_cnt1 <= 2'b00;
              end
            `DDR_WB_WR_DONE:
              begin
                wr_o <= 1'b1;
              end
            `DDR_WB_WR_WAIT:
              begin
                if (pos_gp_cnt1 == 2'b00)
                  begin
                    wr_o <= 1'b1;
                    pos_gp_cnt1 <= pos_gp_cnt1 + 1'b1;
                  end
                else
                  wr_o <= 1'b0;                 
              end
          endcase
        end
    end
  
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          data_avail <= 1'b0;
        end
      else
        begin
          if (wb_adr_i[31:4] == burst_adr)
            begin
              data_avail <= 1'b1;
            end 
          else
            begin
              data_avail <= 1'b0;
            end  
        end  
    end
    
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          wb_ack_o <= 1'b0;
          buf_full <= 1'b0;
          neg_gp_cnt <= 8'h0;       
          data_ready <= 1'b1;
          burst_clear <= 1'b0;
        end
      else
        begin
          case (ddr_wb_fsm_state)         // Synthesis full case
            `DDR_WB_IDLE:
              begin
                wb_ack_o <= 1'b0;
                neg_gp_cnt <= 8'h0;			
                next_state1 <= 1'b0;
                burst_clear <= 1'b0;
                if (wb_cyc_i)
                  mask_reg <= 16'hFFFF;
              end
            `DDR_WB_READ:
              begin
                adr_o <= wb_adr_i;
              end
            `DDR_WB_READ_H:
              begin
                wb_data_o <= data_i;
                wb_ack_o <= 1'b1;     
              end
            `DDR_WB_READ_M:
              begin
                wb_ack_o <= 1'b0;
              end
            `DDR_WB_WRITE:
              begin
                if (wb_cyc_i)
                  begin
                    data_bufw[wb_adr_i[3:2]] <= wb_data_i;
                    case (wb_adr_i[3:2])        // Synthesis full case
                      2'b00: mask_reg <= mask_reg & {12'hFFF,~(wb_sel_i)};
                      2'b01: mask_reg <= mask_reg & {8'hFF,~(wb_sel_i),4'hF};
                      2'b10: mask_reg <= mask_reg & {4'hF,~(wb_sel_i),8'hFF};
                      2'b11: mask_reg <= mask_reg & {~(wb_sel_i),12'hFFF};
                    endcase  
                  end
                else
                  begin
                    next_state1 <= 1'b1;
                    data_ready <= 1'b0; 
                  end
                  
                if (wb_adr_i[3:2] == 2'b11)
                  begin
                    wb_ack_o <= 1'b0;
                    buf_full <= 1'b1;
                  end
                else
                  begin
                    wb_ack_o <= 1'b1;      
                    buf_full <= 1'b0;
                  end
                if (neg_gp_cnt == 8'h0)
                  begin
                    neg_gp_cnt <= neg_gp_cnt + 1'b1;
                    adr_o <= wb_adr_i;      // Need to secure first received address
                  end 
              end    
            `DDR_WB_WR_WAIT:
              begin
                wb_ack_o <= 1'b0;
                if (wr_done_i)
                  begin
                    buf_full <= 1'b0;
                    wb_ack_o <= 1'b1;
                  end
              end
            `DDR_WB_WR_DONE:
              begin
                wb_ack_o <= 1'b0;
                next_state1 <= 1'b0;
                data_ready <= 1'b1;
                burst_clear <= 1'b1;
              end  
          endcase
        end
    end
  
  always @ (posedge clk_div_i)
    begin
      if (!rst_i)
        begin
          buf_rdy <= 1'b1;
        end
      else
        begin
          if (ddr_wb_fsm_state == `DDR_WB_WRITE)
            buf_rdy <= 1'b0;
          else if (wr_done_i)
            buf_rdy <= 1'b1;    
        end        
    end
    
  always @ (negedge clk_div_i)
    begin
      if (!rst_i)
        begin
          burst_adr <= 28'h0;           // At start it should be set to nonexistent address.   
          burst_cnt <= 2'b00;
        end
      else
        begin
          if (rd_done_i)
            begin
              if (burst_cnt != 2'b01)
                burst_cnt <= burst_cnt + 1'b1;
                
              if (burst_cnt == 2'b00)
                burst_adr <= wb_adr_i[31:4];        // This sets flag data_avail
            end
          else
            begin
              burst_cnt <= 2'b00;
              if (burst_clear)
                burst_adr <= 28'h0;
            end
        end
    end      
      
endmodule