//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// UART controller
//
							
  `define UART_IDLE         4'b0000                        // UART controller is in idle state
   
  `define UART_CMD_READ     4'b0001                        // read command word
  `define UART_TLOOP        4'b0010                        // UART controller is running test loop 
  `define UART_READ         4'b0011                        // During this stage data is read from UART
  `define UART_READ1        4'b0100                        // Auxiliary state of FSM  for UART_READ
  `define UART_WBWR         4'b0101                        // Write 32-bit word to the WB
  `define UART_WRITE        4'b0110                        // During this stage data is written to UART
  `define UART_WRITE1       4'b0111                        // Auxiliary state of FSM for UART_WRITE
  `define UART_WRITE2       4'b1000                        // Stage which actually writes data to UART
  `define UART_WRITE3       4'b1001                        // Auxiliary stage for WRITE2
  `define UART_WB_READ      4'b1010                        // State to read WB data in slave mode
  
  // Decoding of PC's commands:
  // UART_TLOOP - one byte of data:
  //        8'h31 - echo back pattern 8'h33 15 times
  //        8'h32 - echo back pattern 8'h45 15 times
  // UART_MWRITE - command to write data to the main memory controlled by wb_host. Format of the command is as following:
  //            !8'h33!32'h(address)!32'h(size in 32bit words)!data!
  // UART_MREAD - command to read data in the main memory and send it via UART. Format of the command is as following:
  //            !8'h34!32'h(address)!32'h(size in 32bit words)!
  //
module uart_cntr (
  clk_i, rst_i, data_i, uart_data_i, uart_txrdy_i,
  uart_rxrdy_i, 
  uart_we_o, uart_rd_o, data_o,
  // Wishbone interface
  wb_data_i, wb_adr_i, wb_we_i, wb_sel_i, wb_stb_i, wb_ack_i,
  wb_cyc_i, 
  wb_data_o, wb_adr_o, wb_we_o, wb_sel_o, wb_stb_o, wb_ack_o,
  wb_cyc_o,
  uart_led_o
  );
  input clk_i;                        // Input clock signal
  input rst_i;                        // Input reset signal
  input [31:0] data_i;                // Input data from wbhost
  input [7:0] uart_data_i;            // Input data from UART
  input uart_txrdy_i;                 // Tx ready signal from UART
  input uart_rxrdy_i;                 // Rx ready signal from UART
  output uart_we_o;                   // Write enable signal for UART
  output uart_rd_o;                   // Read data from UART
  output [7:0] data_o;                // Output data for UART
  input [31:0] wb_data_i;             // Input data from WB
  input [31:0] wb_adr_i;              // Input address from WB
  input wb_we_i;                      // Write enable signal
  input [3:0] wb_sel_i;               // Select signal
  input wb_stb_i;                     // Input strobe
  input wb_ack_i;                     // Acknowledge
  input wb_cyc_i;                     // Cycle
  output [31:0] wb_data_o;            // Output data for the wbhost
  output [31:0] wb_adr_o;             // Output address
  output wb_we_o;                     // Write enable signal
  output [3:0] wb_sel_o;              // Select
  output wb_stb_o;                    // Strobe
  output wb_ack_o;                    // Acknowledge
  output wb_cyc_o;                    // Cycle
  output [3:0] uart_led_o;            // Test output to see on the LEDs acquired word
  
  reg uart_we_o;
  reg uart_rd_o;
  reg [7:0] data_o;
  reg [31:0] wb_data_o;
  reg wb_we_o;
  reg [3:0] wb_sel_o;
  reg wb_stb_o;
  reg wb_ack_o;
  reg wb_cyc_o;
 
  reg next_state1;
  reg next_state2; 
  reg [31:0] count1;
  reg [3:0] count2;
  reg [7:0] uart_data_buf;
  reg uart_rd_dly;
  reg cnt;
  
  reg [3:0] uart_fsm_state;
  reg [3:0] uart_fsm_next_state;
  reg [31:0] mem_adr;
  reg [31:0] data_size;
  reg [31:0] uart_data;
  reg [31:0] wb_mdata;
  reg [2:0] byte_cnt;
  reg wb_ack_pos;
  reg uart_rxrdy_neg;  
  reg [3:0] uart_fsm_state_pos;

  assign wb_adr_o = mem_adr;   
  assign uart_led_o = uart_fsm_state;
//-----------------------------------------UART cntr state machine------------------------------  
  always @ (uart_fsm_state or next_state1 or next_state2 or uart_data_i or wb_ack_pos or uart_rd_dly
            or wb_cyc_i)
    begin
      uart_fsm_next_state = `UART_IDLE;
      
      case (uart_fsm_state)                    // Synthesis full case
        `UART_IDLE:
          begin
            if (next_state1)
              uart_fsm_next_state = `UART_CMD_READ;
            else if (wb_cyc_i)
              uart_fsm_next_state = `UART_WB_READ;
            else
              uart_fsm_next_state = `UART_IDLE;
          end
					   
        `UART_CMD_READ:
          begin
            if (uart_rd_dly)
              begin
                case (uart_data_i)                        // Synthesis full case
                  8'h31, 8'h32: uart_fsm_next_state = `UART_TLOOP;                    // Go to loopback test
                  8'h33: uart_fsm_next_state = `UART_READ;
                  8'h34: uart_fsm_next_state = `UART_WRITE;
                  default: uart_fsm_next_state = `UART_IDLE;
                endcase
              end
            else
              uart_fsm_next_state = `UART_CMD_READ;
          end
        `UART_TLOOP:
          begin
            if (next_state2)
              uart_fsm_next_state = `UART_IDLE;
            else 
              uart_fsm_next_state = `UART_TLOOP;
          end
 
        `UART_READ:
          begin
            if (next_state2)
              uart_fsm_next_state = `UART_IDLE;
            else if (next_state1)
              uart_fsm_next_state = `UART_WBWR;
            else
              uart_fsm_next_state = `UART_READ1;
          end
        `UART_READ1:
          begin
            uart_fsm_next_state = `UART_READ;
          end
        `UART_WBWR:
          begin
            if (wb_ack_i)
              uart_fsm_next_state = `UART_READ;
            else
              uart_fsm_next_state = `UART_WBWR;  
          end
        `UART_WRITE:
          begin
            if (next_state2)
              uart_fsm_next_state = `UART_IDLE;
            else if (next_state1)
              uart_fsm_next_state = `UART_WRITE2;
            else
              uart_fsm_next_state = `UART_WRITE1;
          end
        `UART_WRITE1:
          begin
            uart_fsm_next_state = `UART_WRITE;
          end
        `UART_WRITE2:
          begin
            if (next_state2)
              uart_fsm_next_state = `UART_IDLE;
            else if (wb_ack_pos)
              uart_fsm_next_state = `UART_WRITE3;
            else
              uart_fsm_next_state = `UART_WRITE2;  
          end
        `UART_WRITE3:
          begin
            if (next_state1)
              uart_fsm_next_state = `UART_WRITE2;
            else
              uart_fsm_next_state = `UART_WRITE3;
          end
        `UART_WB_READ:
          begin
            if (next_state2)
              uart_fsm_next_state = `UART_IDLE;
            else
              uart_fsm_next_state = `UART_WB_READ;
          end
      endcase
    end		 
    
   always @ (negedge clk_i)
     begin
       if (~rst_i)
         begin
          uart_fsm_state <= `UART_IDLE;
         end
       else
         begin
           uart_fsm_state <= uart_fsm_next_state;
         end
     end
  
  
  //-------------------------------------End of UART cntr state machine---------------------------
  
  always @ (posedge clk_i)
    begin
      if (~rst_i)
        begin
          uart_we_o <= 1'b0;
          uart_rd_o <= 1'b0;
          next_state1 <= 1'b0;
          next_state2 <= 1'b0;
          wb_cyc_o <= 1'b0;
          wb_we_o <= 1'b0;
          wb_stb_o <= 1'b0;
        end
      else
        begin
          case (uart_fsm_state)                        // Synthesis full case
           `UART_IDLE:
             begin
               count2 <= 4'h0;
               next_state2 <= 1'b0;
               if (uart_rxrdy_neg)
                 begin
                   uart_rd_o <= 1'b1;
                   next_state1 <= 1'b1;
                 end
             end
           `UART_CMD_READ:
             begin
               uart_rd_o <= 1'b0;                // Read only one word during command stage  
               next_state1 <= 1'b0;
               next_state2 <= 1'b0; 
               count1 <= 32'h0;
               count2 <= 4'h0;
               data_size <= 32'h1;
   
             end				
           `UART_TLOOP:
              begin
                if (count1 == 32'h00000002)
                  begin
                    count1 <= 32'h0;
                    uart_we_o <= 1'b1;
                    count2 <= count2 + 1'b1;
                    if (uart_data_buf == 8'h31)
 
                      data_o <= 8'h33;
                    else if (uart_data_buf == 8'h32)
                      data_o <= 8'h45;
                  end						
                else
                  begin
                    count1 <= count1 + 1'b1;
                    uart_we_o <= 1'b0;
                  end
                if (count2 == 4'hF)
                  begin
                    next_state2 <= 1'b1;
                    count1 <= 32'h0;
                  end
                next_state1 <= 1'b0;
              end
	 
            `UART_READ:
              begin
                if (uart_rxrdy_neg)
                  begin
                    uart_rd_o <= 1'b1;
                    count1 <= count1 + 1'b1;
                  end
                if (count1 == 32'h4)
                  mem_adr <= (uart_data - 4'h4);
                else if (count1 == 32'h8)
                  data_size <= uart_data;
                else if ((count1 != 32'h0) && (count1[1:0] == 2'b00) && uart_rd_dly)      // All the rest words are data, which needs to be written over wishbone
                  begin
                    wb_data_o <= uart_data;
                    data_size <= data_size - 1'b1;        // Start to decrement word count  
                    mem_adr <= mem_adr + 4'h4;
                    wb_sel_o <= 4'b1111;
                    {wb_cyc_o, wb_stb_o, wb_we_o} <= 3'b111;
                    next_state1 <= 1'b1;          // Switch to the next stage
                  end
                if (data_size == 32'h0)
                  next_state2 <= 1'b1;
              end
            `UART_READ1:
              begin
                uart_rd_o <= 1'b0;
              end
            `UART_WBWR:
              begin
                next_state1 <= 1'b0;
                next_state2 <= 1'b0;
                if (wb_ack_i)
                  {wb_cyc_o, wb_stb_o, wb_we_o} <= 3'b000;
              end
            `UART_WRITE:
              begin
                next_state2 <= 1'b0;
                cnt <= 2'b00;
                if (uart_rxrdy_neg)
                  begin
                    uart_rd_o <= 1'b1;
                    count1 <= count1 + 1'b1;
                  end
                if (count1 == 32'h4)
                  mem_adr <= uart_data;
                else if (count1 == 32'h8)
                  begin
                    data_size <= uart_data;
                    next_state1 <= 1'b1;
                  end
              end
            `UART_WRITE1:
              begin
                uart_rd_o <= 1'b0;
              end  
            `UART_WRITE2:
              begin
                next_state1 <= 1'b0;
                if (wb_ack_i)
                  begin
                    mem_adr <= mem_adr + 4'h4;
                    wb_mdata <= wb_data_i;
                    data_size <= data_size - 1'b1;
                  end
                if (data_size == 32'h0)
                  next_state2 <= 1'b1;
                else
                  begin
                    wb_sel_o <= 4'b1111;
                    {wb_cyc_o, wb_stb_o, wb_we_o} <= 3'b110; 
                    cnt <= 1'b0;
                    byte_cnt <= 3'b000;
                  end
              end
            `UART_WRITE3:
              begin
                {wb_cyc_o, wb_stb_o, wb_we_o} <= 3'b000;
                cnt <= cnt + 1'b1;
                if (cnt && uart_txrdy_i)
                  begin
                    byte_cnt <= byte_cnt + 1'b1;
                    uart_we_o <= 1'b1;
                    data_o <= wb_mdata[31:24];
                    wb_mdata <= {wb_mdata[23:0],8'h00};
                  end 
                else
                  begin
                    uart_we_o <= 1'b0;
                  end
                if (byte_cnt == 3'b100)
                  next_state1 <= 1'b1;
                else
                  next_state1 <= 1'b0;
              end
            `UART_WB_READ:
              begin
                count2 <= count2 + 1'b1;
                case (count2)           // Synthesis full case
                  4'h0: data_o <= wb_data_i[31:24];
                  4'h1: data_o <= wb_data_i[23:16];
                  4'h2: data_o <= wb_data_i[15:8];
                  4'h3: data_o <= wb_data_i[7:0];
                endcase  
                if (count2 < 4'h4)
                  uart_we_o <= 1'b1;
                else 
                  begin
                    uart_we_o <= 1'b0;
                    next_state2 <= 1'b1;
                  end
              end
          endcase
        end
    end
 
  always @ (posedge clk_i)
    begin
      uart_rd_dly <= uart_rd_o;
      wb_ack_pos <= wb_ack_i; 
      uart_fsm_state_pos <= uart_fsm_state;
    end
    
   always @ (negedge clk_i)
     begin
       if (!rst_i)
         begin
           wb_ack_o <= 1'b0;  
         end
       else
         begin
           case (uart_fsm_state_pos)                        // Synthesis full case
             `UART_IDLE:
               begin
                 wb_ack_o <= 1'b0;
               end
             `UART_CMD_READ:
               begin
                 uart_data_buf <= uart_data_i;
               end
             `UART_READ1, `UART_WRITE1:
               begin
                 if (uart_rd_dly)
                   uart_data <= {uart_data[23:0],uart_data_i};
               end
              `UART_WB_READ:
                begin
                  if (count2 == 4'h3)
                    wb_ack_o <= 1'b1;
                  else
                    wb_ack_o <= 1'b0;
                end
           endcase
         end
       uart_rxrdy_neg <= uart_rxrdy_i;
     end
    
endmodule