//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Wishbone host
//
// Access priorities are as following (from higher priorities to lower):
// - Ethernet controller
// - UART controller
// - DDR3 controller
// - SD card controller
// - CPU
//
`include "wbhost_parms.v"
//
module wb_host (
  clk_i, rst_i,
  // WB signals common to all devices
  wb_adr_o, wb_data_o, wb_we_o, wb_sel_o, wb_stb_o, wb_ack_o,
  // Interface signals with CPU
  cpu_adr_i, cpu_data_i, cpu_we_i, cpu_sel_i, cpu_stb_i, cpu_cyc_i, cpu_ack_i,
  // Interface signals with DDR3 SDRAM controller
  ddr_adr_i, ddr_data_i, ddr_we_i, ddr_sel_i, ddr_stb_i, ddr_cyc_i, ddr_ack_i,
  ddr_cyc_o,
  // Interface signals with UART controller
  uart_adr_i, uart_data_i, uart_we_i, uart_sel_i, uart_stb_i, uart_cyc_i, uart_ack_i,
  uart_cyc_o,
  // Interface signals with Ethernet controller
  eth_adr_i, eth_data_i, eth_we_i, eth_sel_i, eth_stb_i, eth_cyc_i, eth_ack_i,
  eth_cyc_o,
  // Interface signals with CF card controller
  cf_adr_i, cf_data_i, cf_we_i, cf_sel_i, cf_stb_i, cf_cyc_i, cf_ack_i,
  cf_cyc_o
  );
  
  input clk_i;                  // Input clock
  input rst_i;                  // Input reset signal
// WB signals common to all devices
  output [31:0] wb_adr_o;
  output [31:0] wb_data_o;
  output wb_we_o;
  output [3:0] wb_sel_o;
  output wb_stb_o;
  output wb_ack_o;
// Interface with CPU  
  input [31:0] cpu_adr_i;
  input [31:0] cpu_data_i;
  input cpu_we_i;
  input [3:0] cpu_sel_i;
  input cpu_stb_i;
  input cpu_cyc_i;
  input cpu_ack_i;
// Interface with DDR3 SDRAM controller
  input [31:0] ddr_adr_i;
  input [31:0] ddr_data_i;
  input ddr_we_i;
  input [3:0] ddr_sel_i;
  input ddr_stb_i;
  input ddr_cyc_i;
  input ddr_ack_i;
  output ddr_cyc_o;
// Interface with UART controller
  input [31:0] uart_adr_i;
  input [31:0] uart_data_i;
  input uart_we_i;
  input [3:0] uart_sel_i;
  input uart_stb_i;
  input uart_cyc_i;
  input uart_ack_i;
  output uart_cyc_o;
// Interface with Ethernet controller
  input [31:0] eth_adr_i;
  input [31:0] eth_data_i;
  input eth_we_i;
  input [3:0] eth_sel_i;
  input eth_stb_i;
  input eth_cyc_i;
  input eth_ack_i;
  output eth_cyc_o;
// Interface with CF card controller
  input [31:0] cf_adr_i;
  input [31:0] cf_data_i;
  input cf_we_i;
  input [3:0] cf_sel_i;
  input cf_stb_i;
  input cf_cyc_i;
  input cf_ack_i;
  output cf_cyc_o;
  
  // Internal wires and registers
  wire wb_cyc_drv;
  
  //----------------------------WB host state machine----------------------------------
  reg [3:0] wbhost_fsm_state;
  reg [3:0] wbhost_fsm_next_state;
  
  always @ (wbhost_fsm_state or cpu_cyc_i or ddr_cyc_i        // All signals in the sensitivity list change on the posedge 
            or uart_cyc_i or eth_cyc_i or cf_cyc_i or uart_we_i)
    begin
      wbhost_fsm_next_state = `WBHOST_IDLE;
      case (wbhost_fsm_state)                 // Synthesis full case
        `WBHOST_IDLE:
          begin
            if (eth_cyc_i)
              wbhost_fsm_next_state = `WBHOST_ETH;
            else if (uart_cyc_i)
              wbhost_fsm_next_state = `WBHOST_UART;
            else if (ddr_cyc_i)
              wbhost_fsm_next_state = `WBHOST_DDR;
            else if (cf_cyc_i)
              wbhost_fsm_next_state = `WBHOST_CF;
            else if (cpu_cyc_i)
              wbhost_fsm_next_state = `WBHOST_CPU;
            else
              wbhost_fsm_next_state = `WBHOST_IDLE;
          end
        `WBHOST_ETH:
          begin
            if (eth_cyc_i)
              wbhost_fsm_next_state = `WBHOST_ETH;
            else
              wbhost_fsm_next_state = `WBHOST_IDLE;
          end
        `WBHOST_UART:
          begin
            if (uart_cyc_i)
              wbhost_fsm_next_state = `WBHOST_UART;
            else
              wbhost_fsm_next_state = `WBHOST_IDLE;
          end
        `WBHOST_DDR:
          begin
            if (ddr_cyc_i)
              wbhost_fsm_next_state = `WBHOST_DDR;
            else
              wbhost_fsm_next_state = `WBHOST_IDLE;
          end
        `WBHOST_CF:
          begin
            if (cf_cyc_i)
              wbhost_fsm_next_state = `WBHOST_CF;
            else
              wbhost_fsm_next_state = `WBHOST_IDLE;
          end
        `WBHOST_CPU:
          begin
            if (cpu_cyc_i)
              wbhost_fsm_next_state = `WBHOST_CPU;
            else
              wbhost_fsm_next_state = `WBHOST_IDLE;
          end
      endcase
    end
  
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          wbhost_fsm_state <= `WBHOST_IDLE;
        end
      else
        begin
          wbhost_fsm_state <= wbhost_fsm_next_state;
        end
    end
  
  //--------------------------End of WB host state machine-----------------------------

  assign wb_adr_o = (wbhost_fsm_state == `WBHOST_ETH) ? eth_adr_i :
                    (wbhost_fsm_state == `WBHOST_UART) ? uart_adr_i :
                    (wbhost_fsm_state == `WBHOST_DDR) ? ddr_adr_i :
                    (wbhost_fsm_state == `WBHOST_CF) ? cf_adr_i :
                    (wbhost_fsm_state == `WBHOST_CPU) ? cpu_adr_i : 32'hZZZZZZZZ;
  assign wb_data_o = (wbhost_fsm_state == `WBHOST_ETH) ? eth_data_i :
                     ((wbhost_fsm_state == `WBHOST_UART) && uart_we_i) ? uart_data_i :
                     ((wbhost_fsm_state == `WBHOST_UART) && ~uart_we_i && ddr_cyc_o) ? ddr_data_i :
                     (wbhost_fsm_state == `WBHOST_DDR) ? ddr_data_i :
                     (wbhost_fsm_state == `WBHOST_CF) ? cf_data_i :
                     (wbhost_fsm_state == `WBHOST_CPU) ? cpu_data_i : 32'hZZZZZZZZ;
  assign wb_we_o = (wbhost_fsm_state == `WBHOST_ETH) ? eth_we_i :
                   (wbhost_fsm_state == `WBHOST_UART) ? uart_we_i :
                   (wbhost_fsm_state == `WBHOST_DDR) ? ddr_we_i :
                   (wbhost_fsm_state == `WBHOST_CF) ? cf_we_i :
                   (wbhost_fsm_state == `WBHOST_CPU) ? cpu_we_i : 1'b0;
  assign wb_sel_o = (wbhost_fsm_state == `WBHOST_ETH) ? eth_sel_i :
                    (wbhost_fsm_state == `WBHOST_UART) ? uart_sel_i :
                    (wbhost_fsm_state == `WBHOST_DDR) ? ddr_sel_i :
                    (wbhost_fsm_state == `WBHOST_CF) ? cf_sel_i :
                    (wbhost_fsm_state == `WBHOST_CPU) ? cpu_sel_i : 4'b0000;
  assign wb_stb_o = (wbhost_fsm_state == `WBHOST_ETH) ? eth_stb_i :
                    (wbhost_fsm_state == `WBHOST_UART) ? uart_stb_i :
                    (wbhost_fsm_state == `WBHOST_DDR) ? ddr_stb_i :
                    (wbhost_fsm_state == `WBHOST_CF) ? cf_stb_i :
                    (wbhost_fsm_state == `WBHOST_CPU) ? cpu_stb_i : 1'b0;
  assign wb_cyc_drv = (wbhost_fsm_state == `WBHOST_ETH) ? eth_cyc_i :
                      (wbhost_fsm_state == `WBHOST_UART) ? uart_cyc_i :
                      (wbhost_fsm_state == `WBHOST_DDR) ? ddr_cyc_i :
                      (wbhost_fsm_state == `WBHOST_CF) ? cf_cyc_i :
                      (wbhost_fsm_state == `WBHOST_CPU) ? cpu_cyc_i : 1'b0;
                      
  assign wb_ack_o = (wb_adr_o[31] === `BASE_DDR) ? ddr_ack_i :
                    (wb_adr_o[31:12] === `BASE_UART) ? uart_ack_i :
                    (wb_adr_o[31:12] === `BASE_ETH) ? eth_ack_i :
                    (wb_adr_o[31:12] === `BASE_CF) ? cf_ack_i : 1'b0;                      
  assign ddr_cyc_o = (wb_adr_o[31] == `BASE_DDR) ? wb_cyc_drv : 1'b0;
  assign uart_cyc_o = (wb_adr_o[31:12] == `BASE_UART) ? wb_cyc_drv : 1'b0;
  assign eth_cyc_o = (wb_adr_o[31:12] == `BASE_ETH) ? wb_cyc_drv : 1'b0;
  assign cf_cyc_o = (wb_adr_o[31:12] == `BASE_CF) ? wb_cyc_drv : 1'b0; 
  

  
endmodule